/*
 * Copyright (c) 1998-2007 TeamDev Ltd. All Rights Reserved.
 * Use is subject to license terms.
 */

// -------------------------- COMMON UTILITIES

function q__getRowsInSection(table, tableSection) {
  var section = q__getChildNodesWithNames(table, [tableSection])[0];
  var result = (section != null) ? q__getChildNodesWithNames(section, ["tr"]) : [];
  return result;
}

function q__getRowCells(row) {
  var safari = false;
  var cells = !safari ? row.cells : q__getChildNodesWithNames(row, ["td", "th"])
  return cells
}

function q__getCellColSpan(cell) {
  var safari = false; // todo: check under safari when colspans are implemented
  var colSpan = !safari ? cell.colSpan : cell.getAttribute('colSpan');
  if (!colSpan || colSpan == -1)
    colSpan = 1;
  return colSpan;
}

function q__stringEmpty(str) {
  return (!str || str == "");
}

function q__initTableStyles(columns, commonHeaderExists, filterRowExists, commonFooterExists, noDataRows,
                            gridLines, rowStyles, rowStylesMap, cellStylesMap) {
  var table = this;
  table._commonHeaderExists = commonHeaderExists;
  table._commonFooterExists = commonFooterExists;
  table._filterRowExists = filterRowExists;
  table._noDataRows = noDataRows;

  // bodyRowClass, bodyOddRowClass, rolloverRowClass, commonHeaderRowClass, headerRowClass, filterRowClass, commonFooterRowClass, footerRowClass
  (function(bodyRowClass, bodyOddRowClass, rolloverRowClass, commonHeaderRowClass, headerRowClass, filterRowClass, commonFooterRowClass, footerRowClass) {
    table._commonHeaderRowClass = commonHeaderRowClass;
    table._commonFooterRowClass = commonFooterRowClass;
    table._headerRowClass = headerRowClass;
    table._filterRowClass = filterRowClass;
    table._footerRowClass = footerRowClass;
    table._bodyRowClass = bodyRowClass;
    table._bodyOddRowClass = bodyOddRowClass;
    table._rolloverRowClass = rolloverRowClass;
  }).apply(null, rowStyles);
  table._rowStylesMap = rowStylesMap;
  table._cellStylesMap = cellStylesMap;

  // initialize grid lines
  var tempIdx = 0;
  table._horizontalGridLines   = gridLines[tempIdx++];
  table._verticalGridLines     = gridLines[tempIdx++];
  table._commonHeaderSeparator = gridLines[tempIdx++];
  table._commonFooterSeparator = gridLines[tempIdx++];
  table._headerHorizSeparator  = gridLines[tempIdx++];
  table._headerVertSeparator   = gridLines[tempIdx++];
  table._filterRowSeparator    = gridLines[tempIdx++];
  table._footerHorizSeparator  = gridLines[tempIdx++];
  table._footerVertSeparator   = gridLines[tempIdx++];

  table.style.emptyCells = "show";
  if (q__isExplorer()) {
    var bordersNeeded =
            !q__stringEmpty(table._horizontalGridLines)   ||
            !q__stringEmpty(table._verticalGridLines)     ||
            !q__stringEmpty(table._commonHeaderSeparator) ||
            !q__stringEmpty(table._commonFooterSeparator) ||
            !q__stringEmpty(table._headerHorizSeparator)  ||
            !q__stringEmpty(table._headerVertSeparator)   ||
            !q__stringEmpty(table._filterRowSeparator)    ||
            !q__stringEmpty(table._footerHorizSeparator)  ||
            !q__stringEmpty(table._footerVertSeparator);
    if (bordersNeeded)
      table.style.borderCollapse = "collapse";
  }

  q__initTableRows(table);
  q__initTableColumns(table, columns);

  table._removeAllRows = q__table_removeAllRows;
  table._insertRowsAfter = q__table_insertRowsAfter;
  table._rowInsertionCallbacks = new Array();
  table._addRowInsertionCallback = function(callback) {
    table._rowInsertionCallbacks.push(callback);
  }
}

function q__table_removeAllRows() {
  var bodyRows = this._getBodyRows();
  for (var i = 0, count = bodyRows.length; i < count; i++) {
    var row = bodyRows[i];
    row.parentNode.removeChild(row);
  }
  this._bodyRows = undefined;

  this._rowStylesMap = {};
  this._cellStylesMap = {};
}

function q__table_insertRowsAfter(afterIndex, rowsToInsert, newRowsToStylesMap, newRowCellsToStylesMap) {
  var bodyRows = this._getBodyRows();
  var columns = this._columns;
  var afterRow = afterIndex != -1 ? bodyRows[afterIndex] : null;

  var addedRowCount = rowsToInsert.length;
  this._bodyRowCount += addedRowCount;
  for (var originalRowIndex = bodyRows.length - 1; originalRowIndex >= afterIndex && originalRowIndex >= 0; originalRowIndex--) {
    var newRowIndex = originalRowIndex + addedRowCount;
    var moveRow = originalRowIndex != afterIndex;
    var movedRow = bodyRows[originalRowIndex];

    if (moveRow) {
      bodyRows[newRowIndex] = movedRow;
      bodyRows[originalRowIndex] = null;
      movedRow._index = newRowIndex;
    }

    q__calculateInitialRowClass(movedRow, this, moveRow ? newRowIndex : originalRowIndex);

    if (moveRow) {
      this._rowStylesMap[newRowIndex] = this._rowStylesMap[originalRowIndex];
      this._rowStylesMap[originalRowIndex] = null;
    }

    for (var colIndex = 0, colCount = columns.length; colIndex < colCount; colIndex++) {
      var column = columns[colIndex];
      var bodyCells = column._bodyCells;
      var movedCell = bodyCells[originalRowIndex];

      if (moveRow) {
        bodyCells[newRowIndex] = movedCell;
        bodyCells[originalRowIndex] = null;
        var originalCellKey = originalRowIndex + "x" + colIndex;
        var newCellKey = newRowIndex + "x" + colIndex;
        this._cellStylesMap[newCellKey] = this._cellStylesMap[originalCellKey];
        this._cellStylesMap[originalCellKey] = null;
      }

      q__initColumnBodyCellStyles(movedCell);
    }

  }

  var nextNode;
  var rowContainer;
  if (afterRow) {
    nextNode = afterRow.nextSibling;
    rowContainer = afterRow.parentNode;
  } else {
    var bodySection = q__getChildNodesWithNames(this, ["tbody"])[0];
    nextNode = bodyRows.length > 0 ? bodyRows[0] : null;
    rowContainer = bodySection;
  }

  for (var i = 0, count = rowsToInsert.length; i < count; i++) {
    var newRow = rowsToInsert[i];
    if (nextNode != null)
      rowContainer.insertBefore(newRow, nextNode);
    else
      rowContainer.appendChild(newRow);

    var newRowIndex = afterIndex + 1 + i;
    bodyRows[newRowIndex] = newRow;
    this._rowStylesMap[newRowIndex] = newRowsToStylesMap ? newRowsToStylesMap[i] : undefined;
    q__initTableRow(newRow, this, newRowIndex);

    for (var colIndex = 0, colCount = columns.length; colIndex < colCount; colIndex++) {
      var column = columns[colIndex];
      var cellStyleKey = newRowIndex + "x" + colIndex;
      this._cellStylesMap[cellStyleKey] = newRowCellsToStylesMap ? newRowCellsToStylesMap[i + "x" + colIndex] : undefined;
      var cell = q__initTableColumnCell(newRow, column);
      var bodyCells = column._bodyCells;
      bodyCells[newRowIndex] = cell;
      if (cell)
        q__initColumnBodyCell(cell);
    }
  }

  for (var i = 0, count = this._rowInsertionCallbacks.length; i < count; i++) {
    this._rowInsertionCallbacks[i](this, afterIndex, rowsToInsert);
  }


}


// -------------------------- TABLE ROWS SUPPORT

function q__initTableRows(table) {
  //// _headerFacetUsed, _footerFacetUsed
  //// _noDataRows, _filterRowUsed

  //_filterRowClass, _headerRowClass
  // _commonHeaderRowClass,
  // _bodyRowClass, _bodyOddRowClass, _rolloverRowClass
  // _commonFooterRowClass, _footerRowClass;
  // _rowStylesMap, _cellStylesMap

  table._getHeaderRows = function() {
    if (!this._headerRows)
      this._headerRows = q__getRowsInSection(this, "thead");
    return this._headerRows;
  }
  table._getBodyRows = function() {
    if (!this._bodyRows) {
      this._bodyRows = q__getRowsInSection(this, "tbody");
      this._bodyRowCount = this._bodyRows.length;
    }
    return this._bodyRows;
  }
  table._getFooterRows = function() {
    if (!this._footerRows)
      this._footerRows = q__getRowsInSection(this, "tfoot");
    return this._footerRows;
  }

  var headRows = table._getHeaderRows();
  for (var rowIndex = 0, rowCount = headRows.length; rowIndex < rowCount; rowIndex++) {
    var row = headRows[rowIndex];
    q__initRowCells(row);
  }
  var commonHeaderRowIndex = table._commonHeaderExists ? 0 : -1;
  var columnHeadersRowIndex = table._commonHeaderExists ? 1 : 0;
  var filterRowIndex = table._filterRowExists ? headRows.length - 1 : -1;
  table._filterRowIndex = filterRowIndex;
  table._columnHeadersRowIndex = columnHeadersRowIndex;
  table._commonHeaderRowIndex = commonHeaderRowIndex;
  if (filterRowIndex != -1 && columnHeadersRowIndex == filterRowIndex)
    columnHeadersRowIndex = -1;
  if (columnHeadersRowIndex >= headRows.length) {
    columnHeadersRowIndex = -1;
  }

  if (commonHeaderRowIndex != -1)
    q__appendRowCellClassNames(headRows[commonHeaderRowIndex], [table._commonHeaderRowClass]);
  if (filterRowIndex != -1)
    q__appendRowCellClassNames(headRows[filterRowIndex], [table._filterRowClass]);
  if (columnHeadersRowIndex != -1)
    q__appendRowCellClassNames(headRows[columnHeadersRowIndex], [table._headerRowClass]);

  var bodyRows = table._getBodyRows();
  for (var rowIndex = 0, rowCount = bodyRows.length; rowIndex < rowCount; rowIndex++) {
    var row = bodyRows[rowIndex];
    q__initTableRow(row, table, rowIndex);
  }
  var footRows = table._getFooterRows();
  for (var rowIndex = 0, rowCount = footRows.length; rowIndex < rowCount; rowIndex++) {
    var row = footRows[rowIndex];
    q__initRowCells(row);
  }
  var footerRowIndex = footRows.length - 1;
  if (table._commonFooterExists)
    q__appendRowCellClassNames(footRows[footerRowIndex--], [table._commonFooterRowClass]);
  if (footerRowIndex >= 0)
    q__appendRowCellClassNames(footRows[footerRowIndex], [table._footerRowClass]);
}

function q__appendRowCellClassNames(row, classNames) {
  var cells = row._cells;
  for (var i = 0, count = cells.length; i < count; i++) {
    var cell = cells[i];
    q__appendCellClassNames(cell, classNames);
  }
}

function q__initTableRow(row, table, rowIndex) {
  row._table = table;
  row._index = rowIndex;
  row._selected = false;
  row._visible = true;

  row._originalClassName = row.className;

  row._mouseIsOver = false;
  if (!table._noDataRows && table._rolloverRowClass) {
    q__addEventHandlerSimple(row, "mouseover", "_mouseOverHandler");
    q__addEventHandlerSimple(row, "mouseout", "_mouseOutHandler");
  }
  q__calculateInitialRowClass(row, table, rowIndex);

  if (!row._initialized) {
    row._initialized = true;

    row._isVisible = function () {
      return this._visible;
    }
    row._setVisible = function (visible) {
      if (this._visible == visible)
        return;
      this.style.display = visible ? "" : "none";
      this._visible = visible;
    }

    q__initRowCells(row);

    row._mouseOverHandler = function() {
      this._mouseIsOver = true;
      this._updateStyle();
    }

    row._mouseOutHandler = function() {
      this._mouseIsOver = false;
      this._updateStyle();
    }

    row._updateStyle = function() {
      var cells = this._cells;
      var rowTable = this._table;
      var addedClassName = q__combineClassNames([
              this._selected ? rowTable._selectionClass : null,
              this._mouseIsOver ? rowTable._rolloverRowClass : null]);
      if (row._addedClassName == addedClassName)
        return;
      row._addedClassName = addedClassName;
      var opera = q__isOpera();
      for (var i = 0, count = cells.length; i < count; i++) {
        var cell = cells[i];
        var cls = q__combineClassNames([cell._originalClassName, addedClassName]);
        if (cell.className != cls)
          cell.className = cls;
        q__updateCellWrapperStyle(cell);
        if (opera) {
          var oldBackground = cell.style.background;
          cell.style.background = "white";
          cell.style.background = "#fefefe";
          cell.style.background = oldBackground;
        }
      }
    }
  }

}


function q__initRowCells(row) {
  var cells = q__getRowCells(row);
  row._cells = cells;
  row._cellsByColumns = new Array();
  var physicalColIndex = 0;
  for (var cellIndex = 0, cellCount = cells.length; cellIndex < cellCount; cellIndex++) {
    var cell = cells[cellIndex];
    cell._row = row;
    var cellColSpan = q__getCellColSpan(cell);
    cell._colSpan = cellColSpan;
    cell._physicalColIndex = physicalColIndex;
    row._cellsByColumns[physicalColIndex] = cell;
    physicalColIndex += cellColSpan;
    if (cell.innerHTML == "")
      cell.innerHTML = "&#160;";
  }
  row._getCellByPhysicalColIndex = function(index) {
    var cell = this._cellsByColumns[index];
    return cell ? cell : null;
  }
}

function q__updateCellWrapperStyle(cell) {
  if (cell._noCellWrapperFound)
    return;
  var cellWrapper = cell._cellWrapper;
  if (!cellWrapper) {
    var nodes = q__findChildNodesByClass(cell, "q_cellWrapper", true);
    if (nodes.length == 0) {
      cell._noCellWrapperFound = true;
      return;
    }
    q__assert(nodes.length == 1, "q__updateCellWrapperStyle: no more than one cellwrapper expected, but found: " + nodes.length);
    cellWrapper = nodes[0];
    q__assert(cellWrapper, "q__updateCellWrapperStyle: non-null cellWrapper expected");
    cell._cellWrapper = cellWrapper;
  }
  if (cellWrapper.className != cell.className)
    cellWrapper.className = cell.className;
  cellWrapper.style.width = "100%";
}


function q__calculateInitialRowClass(row, table, rowIndex) {
  var rowClass;
  if (!table._noDataRows)
    rowClass = (rowIndex % 2 == 0) ? table._bodyRowClass : (table._bodyOddRowClass ? table._bodyOddRowClass : table._bodyRowClass);
  else
    rowClass = null;

  row._initialClass = rowClass;
  row._addedClassName = undefined;
}


// -------------------------- TABLE COLUMNS SUPPORT

function q__initTableColumns(table, columns) {
  table._columns = new Array();
  var physicalColIndex = 0;
  var colCount = columns.length;
  table._columnCount = colCount;
  for (var logicalColIndex = 0; logicalColIndex < colCount; logicalColIndex++) {
    var colDataArray = columns[logicalColIndex];
    var i = 0;
    var generalStyleClass  = colDataArray[i++];
    var headerStyleClass   = colDataArray[i++];
    var filterCellStyleClass = colDataArray[i++];
    var bodyStyleClass     = colDataArray[i++];
    var footerStyleClass   = colDataArray[i++];
    var generalEventsArray = colDataArray[i++];
    var headerEventsArray  = colDataArray[i++];
    var bodyEventsArray    = colDataArray[i++];
    var footerEventsArray  = colDataArray[i++];
    var subColCount        = colDataArray[i++];

    var column = new q__tableColumn(table, logicalColIndex, physicalColIndex, subColCount);

    column._generalStyleClass = generalStyleClass;
    column._bodyStyleClass = bodyStyleClass;
    column._generalEventsArray = generalEventsArray;
    column._bodyEventsArray = bodyEventsArray;

    table._columns[logicalColIndex] = column;
    column._subColCount = subColCount;

    var headerCells = column._headerCells;
    if (table._commonHeaderExists) {
      var cell = headerCells[0];
      q__setCellStyleProperty(cell, "borderBottom", q__nullToEmptyStr(table._commonHeaderSeparator));
    }
    var startIndex = table._commonHeaderExists ? 1 : 0;
    for (var cellIndex = startIndex, cellCount = headerCells.length; cellIndex < cellCount; cellIndex++) {
      var cell = headerCells[cellIndex];
      if (cell) {
        if (cellIndex == cellCount - 1)
          q__setCellStyleProperty(cell, "borderBottom", q__nullToEmptyStr(table._headerHorizSeparator != null ? table._headerHorizSeparator : table._horizontalGridLines));
        if (cellIndex == table._filterRowIndex - 1 && table._columnHeadersRowIndex != -1)
          q__setCellStyleProperty(cell, "borderBottom", q__nullToEmptyStr(table._filterRowSeparator));
        if (logicalColIndex != colCount - 1)
          q__setCellRightBorder(cell, q__nullToEmptyStr(table._headerVertSeparator != null ? table._headerVertSeparator : table._verticalGridLines));
        if (cellIndex == startIndex)
          q__appendCellClassNames(cell, [generalStyleClass, headerStyleClass]);
        else
          q__appendCellClassNames(cell, [generalStyleClass, filterCellStyleClass]);
        q__initColumnCell(cell, generalEventsArray, headerEventsArray);
      }
    }

    var bodyCells = column._bodyCells;
    for (var bodyCellIndex = 0, cellCount = bodyCells.length; bodyCellIndex < cellCount; bodyCellIndex++) {
      var cell = bodyCells[bodyCellIndex];
      if (!cell)
        continue;

      q__initColumnBodyCell(cell);
    }

    var footerCells = column._footerCells;
    if (table._commonFooterExists) {
      var cell = footerCells[footerCells.length - 1];
      if (cell)
        q__setCellStyleProperty(cell, "borderTop", q__nullToEmptyStr(table._commonFooterSeparator));
    }
    for (var cellIndex = 0, cellCount = table._commonFooterExists ? footerCells.length - 1 : footerCells.length; cellIndex < cellCount; cellIndex++) {
      var cell = footerCells[cellIndex];
      if (cell) {
        if (cellIndex == 0)
          q__setCellStyleProperty(cell, "borderTop", q__nullToEmptyStr(table._footerHorizSeparator != null ? table._footerHorizSeparator : table._horizontalGridLines));
        if (logicalColIndex != colCount - 1)
          q__setCellRightBorder(cell, q__nullToEmptyStr(table._footerVertSeparator != null ? table._footerVertSeparator : table._verticalGridLines));
        q__appendCellClassNames(cell, [generalStyleClass, footerStyleClass]);
        q__initColumnCell(cell, column._generalEventsArray, footerEventsArray);
      }
    }
    if (subColCount == 0)
      physicalColIndex++;
    else
      physicalColIndex += subColCount;
  }
}

/*
  There are "logical" and "physical" columns. This separation is needed for TreeTable which simulates TreeColumn
  with several columns. Logical columns are columns from the user's point of view, i.e. TreeColumns is a single logical
  column, though it consists of several physical columns. Physical columns are actual columns in the resulting DOM.
 */
function q__tableColumn(table, logicalColIndex, physicalColIndex, subColCount) {
  this._table = table;
  this._logicalColIndex = logicalColIndex;
  this._physicalColIndex = physicalColIndex;
  this._subColCount = subColCount;

  this._headerCells = new Array();
  var headRows = table._getHeaderRows();
  for (var i = 0, count = headRows.length; i < count; i++) {
    var row = headRows[i];
    var cell = row._getCellByPhysicalColIndex(physicalColIndex);
    if (cell) {
      cell._column = this;
      cell._logicalColIndex = logicalColIndex;
      cell._subCells = q__getSubCellsArray(cell, physicalColIndex, subColCount);
    }
    this._headerCells[i] = cell;
  }

  this._bodyCells = new Array();
  var bodyRows = table._getBodyRows();
  for (var i = 0, count = bodyRows.length; i < count; i++) {
    var row = bodyRows[i];
    var cell = q__initTableColumnCell(row, this);
    this._bodyCells[i] = cell;
  }

  this._footerCells = new Array();
  var footRows = table._getFooterRows();
  for (var i = 0, count = footRows.length; i < count; i++) {
    var row = footRows[i];
    var cell = row._getCellByPhysicalColIndex(physicalColIndex);
    if (cell) {
      cell._column = this;
      cell._logicalColIndex = logicalColIndex;
      cell._subCells = q__getSubCellsArray(cell, physicalColIndex, subColCount);
    }
    this._footerCells[i] = cell;
  }
}

function q__setCellStyleProperty(cell, propertyName, propertyValue) {
  if (!cell || !propertyValue)
    return;
  if (!cell._subCells) {
    try {
      cell.style[propertyName] = propertyValue;
    } catch (e) {
      q__logError("q__setCellStyleProperty: couldn't set style property \"" + propertyName + "\" to \"" + propertyValue + "\" ; original error: " + e.message);
      throw e;
    }
    return;
  }

  for (var i = 0, count = cell._subCells.length; i < count; i++) {
    var subCell = cell._subCells[i];
    if (subCell)
      subCell.style[propertyName] = propertyValue;
  }

}

function q__setCellRightBorder(cell, borderValue) {
  if (!cell)
    return;
  if (!cell._subCells) {
    try {
      cell.style.borderRight = borderValue;
    } catch (e) {
      q__logError("q__setCellRightBorder: invalid borderValue: \"" + borderValue + "\" ; it must be a valid CSS declaration of form \"1px solid gray\" ; original error: " + e.message);
      throw e;
    }
    return;
  }
  var lastSubCell = null;
  for (var i = 0, count = cell._subCells.length; i < count; i++) {
    var subCell = cell._subCells[i];
    if (subCell)
      lastSubCell = subCell;
  }
  try {
    lastSubCell.style.borderRight = borderValue;
  } catch (e) {
    q__logError("q__setCellRightBorder: invalid borderValue: " + borderValue + " ; it must be a valid CSS declaration of form '1px solid gray' ; original error: " + e.message);
    throw e;
  }
}

function q__appendCellClassNames(cell, classNames) {
  if (!cell._subCells) {
    q__appendClassNames(cell, classNames);
    q__updateCellWrapperStyle(cell);
    cell._originalClassName = cell.className;
    return;
  }
  var subCells = cell._subCells
  for (var i = 0, count = subCells.length; i < count; i++) {
    var subCell = subCells[i];
    if (!subCell)
      continue;
    subCell._originalClassName = subCell.className;
    q__appendClassNames(subCell, classNames);
    q__updateCellWrapperStyle(cell);
  }
}

function q__initColumnCell(cell, events1, events2) {

  if (!cell._subCells) {
    q__assignEvents(cell, events1);
    q__assignEvents(cell, events2);
  } else {
    var subCells = cell._subCells;
    for (var i = 0, count = subCells.length; i < count; i++) {
      var subCell = subCells[i];
      if (!subCell)
        continue;
      q__assignEvents(subCell, events1);
      q__assignEvents(subCell, events2);
    }
  }
}

function q__initColumnBodyCell(cell) {
  cell._classNameBeforeInitialization = cell.className;
  q__initColumnBodyCellStyles(cell);

  var column = cell._column;
  q__initColumnCell(cell, column._generalEventsArray, column._bodyEventsArray);
}

function q__initColumnBodyCellStyles(cell) {
  cell.className = cell._classNameBeforeInitialization;
  var column = cell._column;
  var row = cell._row;
  var table = row._table;
  var rowIndex = row._index;
  var colIndex = column._logicalColIndex;
  var rowCount = table._bodyRowCount;
  var colCount = table._columnCount;


  var cellKey = rowIndex + "x" + colIndex;
  var individualCellStyle = table._cellStylesMap[cellKey];
  var individualRowClass = table._rowStylesMap[row._index];
  if (table._noDataRows)
    individualCellStyle = null;
  q__appendCellClassNames(cell, [row._initialClass, individualRowClass, individualCellStyle]);

  if (rowIndex != rowCount - 1)
    q__setCellStyleProperty(cell, "borderBottom", q__nullToEmptyStr(table._horizontalGridLines));
  if (colIndex != colCount - 1)
    q__setCellRightBorder(cell, q__nullToEmptyStr(table._verticalGridLines));
  if (!table._noDataRows)
    q__appendCellClassNames(cell, [column._generalStyleClass, column._bodyStyleClass]);
}


function q__assignEvents(element, eventsArray) {
  if (!eventsArray || eventsArray.length == 0)
    return;
  for (var i = 0, count = eventsArray.length; i < count; i++) {
    var eventEntry = eventsArray[i];
    var eventName = eventEntry[0];
    var eventScript = eventEntry[1];

    var handlerFunction;
    eval('handlerFunction = function(event) {return eval(arguments.callee.prototype._handlerScript); }');
    handlerFunction.prototype._handlerScript = eventScript;
    q__addEventHandler(element, eventName, handlerFunction);
  }
}


function q__getSubCellsArray(cell, physicalColIndex, subColCount) {
  var row = cell._row;
  if (subColCount == 0)
    return null;
  var cells = new Array();
  for (var i = 0; i < subColCount; i++) {
    var subCell = row._getCellByPhysicalColIndex(physicalColIndex + i);
    cells.push(subCell);
  }
  return cells;
}

function q__initTableColumnCell(row, column) {
  var colIndex = column._physicalColIndex;
  var cell = row._getCellByPhysicalColIndex(colIndex);
  if (cell) {
    cell._column = column;
    cell._logicalColIndex = column._logicalColIndex;
    cell._subCells = q__getSubCellsArray(cell, column._physicalColIndex, column._subColCount);
  }
  return cell;
}
